// Test program to convert PNGs to a polygonal image format

#define UInt32 unsigned int)(unsigned char
#define UInt16 unsigned short)(unsigned char
#define READINT32(pointer) (((UInt32)*pointer) | ((UInt32)(*(pointer+1))<<8) | ((UInt32)(*(pointer+2))<<16) | ((UInt32)(*(pointer+3))<<24))
#define READINT16(pointer) (((UInt16)*pointer) | ((UInt16)(*(pointer+1))<<8))

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <vector>
#include <sstream>
#include <string>
#include "zlib.h"

// This should help with detecting if 16-bit is viable.
static unsigned char ColorTable5Bit[32] = {
	0, 8, 17, 25, 33, 41, 50, 58, 66, 74, 83, 91, 99,
	107, 116, 124, 132, 140, 149, 157, 165, 173, 182, 190, 198,
	206, 215, 223, 231, 239, 248, 255
};

static unsigned char ColorTable6Bit[64] = {
	0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 45, 49, 53, 57, 61,
	65, 69, 73, 77, 81, 85, 89, 93, 97, 101, 105, 109, 113, 117,
	121, 125, 130, 134, 138, 142, 146, 150, 154, 158, 162, 170,
	174, 178, 182, 190, 194, 198, 202, 206, 210, 215, 219, 223,
	227, 231, 235, 239, 243, 247, 251, 255
};

double round(double decimal)
{
	double intPart = 0.0;
	if(modf(decimal, &intPart) >= 0.5)
		intPart += 1.0;
	return intPart;
}

void Error(const char* error);

/**
 * Data structure to store 1 Pixel of information.
 */
struct Pixel
{
	public:
		Pixel() : a(0), r(0), g(0), b(0)
		{
		}
		Pixel(unsigned char a, unsigned char r, unsigned char g, unsigned char b) : a(a), r(r), g(g), b(b)
		{
		}
		Pixel(unsigned int rgba)
		{
			a = (rgba>>24) & 0xFF;
			r = rgba & 0xFF;
			g = (rgba>>8) & 0xFF;
			b = (rgba>>16) & 0xFF;
		}
		Pixel(const Pixel &px)
		{
			a = px.a;
			r = px.r;
			g = px.g;
			b = px.b;
		}

		bool operator== (Pixel &px)
		{
			return (px.a == a && px.r == r && px.g == g && px.b == b);
		}
		bool operator!= (Pixel &px)
		{
			return (px.a != a || px.r != r || px.g != g || px.b != b);
		}

		unsigned char	a;
		unsigned char	r;
		unsigned char	g;
		unsigned char	b;
};

#define CHUNK 1024
#define COMPRESSION_LEVEL 9
void deflateChunk(const char* data, unsigned int length, char* buffer, unsigned int &size)
{
	unsigned int pos = 0;
	int ret, flush;
	unsigned int have;
	z_stream strm;
	unsigned char in[length];
	memcpy(in, data, length);
	unsigned char out[CHUNK];
	strm.zalloc = Z_NULL;
	strm.zfree = Z_NULL;
	strm.opaque = Z_NULL;
	ret = deflateInit(&strm, COMPRESSION_LEVEL);
	if(ret != Z_OK)
		Error("Unable to init zlib.\n");
	//do compression
	{
		strm.avail_in = length;
		flush = Z_FINISH;
		strm.next_in = in;
		do
		{
			strm.avail_out = CHUNK;
			strm.next_out = out;
			ret = deflate(&strm, flush);
			assert(ret != Z_STREAM_ERROR);
			have = CHUNK - strm.avail_out;
			memcpy(buffer+pos, out, have);
			pos += have;
		}
		while(strm.avail_out == 0);
		assert(strm.avail_in == 0);
	}
	deflateEnd(&strm);
	size = pos;
}

#define MODE_8BIT_PALETTE 0x00
#define MODE_8BIT_GRAYSCALE 0x01
#define MODE_16BIT_COLOR 0x02
#define MODE_8BIT_PALETTE_ALPHA 0x03
#define MODE_8BIT_GRAYSCALE_ALPHA 0x04
#define MODE_24BIT_COLOR 0x05
#define MODE_16BIT_COLOR_ALPHA 0x06
#define MODE_32BIT 0x07

unsigned char CalculateMode(unsigned int depth, bool grayScale, bool alpha)
{
	if(depth == 8)
	{
		if(grayScale)
		{
			if(alpha)
				return MODE_8BIT_GRAYSCALE_ALPHA;
			return MODE_8BIT_GRAYSCALE;
		}
		if(alpha)
			return MODE_8BIT_PALETTE_ALPHA;
		return MODE_8BIT_PALETTE;
	}
	else if(depth == 16)
	{
		if(alpha)
			return MODE_16BIT_COLOR_ALPHA;
		return MODE_16BIT_COLOR;
	}
	else if(depth == 24)
	{
		if(alpha)
			return MODE_32BIT;
		return MODE_24BIT_COLOR;
	}
}

// Returns bytes per pixel
unsigned char GetModeSize(unsigned char mode)
{
	switch(mode)
	{
		case MODE_8BIT_PALETTE:
		case MODE_8BIT_GRAYSCALE:
			return 1;
		case MODE_16BIT_COLOR:
		case MODE_8BIT_PALETTE_ALPHA:
		case MODE_8BIT_GRAYSCALE_ALPHA:
			return 2;
		case MODE_24BIT_COLOR:
		case MODE_16BIT_COLOR_ALPHA:
			return 3;
		case MODE_32BIT:
		default:
			return 4;
	}
}

bool ModeHasAlpha(unsigned char mode)
{
	switch(mode)
	{
		case MODE_8BIT_PALETTE:
		case MODE_8BIT_GRAYSCALE:
		case MODE_16BIT_COLOR:
		case MODE_24BIT_COLOR:
			return false;
		default:
			return true;
	}
}

/**
 * A Post is one line of pixels in a column.  It contains a y-offset and then
 * the length of the line.
 */
struct Post
{
	public:
		Post(unsigned short yOffset) : yOffset(yOffset)
		{
		}

		void Write(std::stringstream &chunk, unsigned char mode, const std::vector<Pixel> &palette)
		{
			unsigned short length = data.size();
			int bpp = GetModeSize(mode);
			char* out = new char[length*bpp + 4];
			out[0] = yOffset & 0xFF;
			out[1] = (yOffset>>8) & 0xFF;
			out[2] = length & 0xFF;
			out[3] = (length>>8) & 0xFF;
			for(unsigned int i = 0;i < length;i++)
			{
				int off = 4;
				if(mode == MODE_24BIT_COLOR || mode == MODE_32BIT)
				{
					out[(i*bpp)+off] = data[i].r;
					out[(i*bpp)+off+1] = data[i].g;
					out[(i*bpp)+off+2] = data[i].b;
					off += 3;
				}
				else if(mode == MODE_16BIT_COLOR || mode == MODE_16BIT_COLOR_ALPHA)
				{
					out[(i*bpp)+off] = static_cast<int> (round(double(data[i].r) / (double(255) / double(31))));
					//bmps use 5-bit green
					out[(i*bpp)+off] |= (static_cast<int> (round(double(data[i].g) / (double(255) / double(31))))<<6) & 0xC0;
					out[(i*bpp)+off+1] = (static_cast<int> (round(double(data[i].g) / (double(255) / double(31))))>>2) & 0x07;
					out[(i*bpp)+off+1] |=(static_cast<int> (round(double(data[i].b) / (double(255) / double(31))))<<3) & 0xF8;
					off += 2;
				}
				else if(mode == MODE_8BIT_GRAYSCALE || mode == MODE_8BIT_GRAYSCALE_ALPHA)
				{
					out[(i*bpp)+off] = data[i].r;
					off++;
				}
				else //8-bit palette
				{
					bool found = false;
					for(unsigned int j = 0;j < palette.size();j++)
					{
						if(palette[j].r == data[i].r && palette[j].g == data[i].g && palette[j].b == data[i].b)
						{
							found = true;
							out[(i*bpp)+off] = j;
							off++;
						}
					}
					if(!found)
						Error("Could not find color in palette!\n");
				}
				if(ModeHasAlpha(mode))
					out[(i*bpp)+off] = data[i].a;
			}
			chunk.write(out, length*bpp + 4);
			delete[] out;
		}

		unsigned short		yOffset;
		std::vector<Pixel>	data;
};

/**
 * This is mainly a container for multiple posts.  Of course it also appends 
 * 0xFFFF when writing.
 */
struct Column
{
	private:
		Column(unsigned char mode) : mode(mode), position(0)
		{
		}

	public:
		~Column()
		{
			for(unsigned int i = 0;i < posts.size();i++)
				delete posts[i];
			posts.clear();
		}

		/// Returns true if the two columns are identical.
		bool Compare(Column *c2)
		{
			if(c2->posts.size() != posts.size())
				return false;
			for(unsigned int i = 0;i < posts.size();i++)
			{
				if(c2->posts[i]->data.size() != posts[i]->data.size())
					return false;
				for(unsigned int j = 0;j < posts[i]->data.size();j++)
				{
					if(c2->posts[i]->data[j] != posts[i]->data[j])
						return false;
				}
			}
			return true;
		}
		/// Takes an array of pixels and splits them into posts.
		static Column *CreateColumn(Pixel* data, unsigned short height, unsigned char mode)
		{
			Column *column = new Column(mode);
			Post *post = NULL;
			for(unsigned short i = 0;i < height;i++)
			{
				if(data[i].a == 0x00)
				{
					if(post != NULL)
					{
						column->posts.push_back(post);
						post = NULL;
					}
					continue;
				}
				if(post == NULL)
				{
					post = new Post(i);
				}
				post->data.push_back(data[i]);
			}
			if(post != NULL)
			{
				column->posts.push_back(post);
			}
			return column;
		}
		unsigned int GetSize()
		{
			unsigned int length = 2;
			for(unsigned int i = 0;i < posts.size();i++)
			{
				length += 4 + (posts[i]->data.size()*GetModeSize(mode));
			}
			return length;
		}
		void Write(std::stringstream &chunk, unsigned char mode, const std::vector<Pixel> &palette)
		{
			for(unsigned int i = 0;i < posts.size();i++)
				posts[i]->Write(chunk, mode, palette);

			char terminator[] = {0xFF, 0xFF};
			chunk.write(terminator, 2);
		}

		std::vector<Post *>	posts;

		unsigned int		position;
		unsigned char		mode;
};

/// New polygonal graphics format.
class PolygonalImage
{
	public:
		PolygonalImage(unsigned short width, unsigned short height, unsigned char mode) : width(width), height(height), mode(mode)
		{
			columnMap = new unsigned int[width];
			for(unsigned short i = 0;i < width;i++)
				columnMap[i] = i;
		}
		~PolygonalImage()
		{
			delete[] columnMap;
		}

		void Write(FILE *file)
		{
			char* out = new char[21];
			out[0] = 'P'; out[1] = 'G'; out[2] = 'F'; out[3] = 0x00;
			out[4] = width & 0xFF;
			out[5] = (width>>8) & 0xFF;
			out[6] = height & 0xFF;
			out[7] = (height>>8) & 0xFF;
			out[8] = mode;
			out[9] = 'I'; out[10] = 'D'; out[11] = 'A'; out[12] = 'T';

			//form the chunk
			std::stringstream chunk;
			unsigned int size = width*4;
			unsigned int removedColumns = 0;
			for(unsigned int i = 0;i < columns.size();i++)
			{
				// Check to see if there is a previous identical column.
				// if so then make this column that column.
				for(unsigned int j = 0;j < i;j++)
				{
					if(columns[i]->Compare(columns[j]))
					{
						removedColumns++;
						delete columns[i];
						columns[i] = columns[j];
						columnMap[i] = columnMap[j];
						break;
					}
				}
				// if the position has been previously set use that and not
				// the size.
				if(columns[i]->position == 0)
				{
					columns[i]->position = size;
					size += columns[i]->GetSize();
				}
				char pointer[4] = {
					(columns[i]->position & 0xFF),
					((columns[i]->position>>8) & 0xFF),
					((columns[i]->position>>16) & 0xFF),
					((columns[i]->position>>24) & 0xFF)
				};
				chunk.write(pointer, 4);
			}
			printf("%d columns removed!\n", removedColumns);
			// Lets calculate the palette.
			if(mode == MODE_8BIT_PALETTE || mode == MODE_8BIT_PALETTE_ALPHA)
			{
				for(unsigned int i = 0;i < columns.size();i++)
				{
					for(unsigned int j = 0;j < columns[i]->posts.size();j++)
					{
						for(unsigned int k = 0;k < columns[i]->posts[j]->data.size();k++)
						{
							bool found = false;
							// and the loops keep going and going and . . . 
							for(unsigned int l = 0;l < palette.size();l++)
							{
								if(palette[l] == columns[i]->posts[j]->data[k])
								{
									found = true;
									break;
								}
							}
							if(!found)
								palette.push_back(Pixel(columns[i]->posts[j]->data[k]));
						}
					}
				}
			}
			for(unsigned int i = 0;i < columns.size();i++)
			{
				if(i == columnMap[i])
					columns[i]->Write(chunk, mode, palette);
			}
			std::string chunkData = chunk.str();
			char* deflatedData = new char[chunkData.length()];
			unsigned int compressedSize = 0;
			deflateChunk(chunkData.c_str(), chunkData.length(), deflatedData, compressedSize);

			out[13] = compressedSize & 0xFF;
			out[14] = (compressedSize>>8) & 0xFF;
			out[15] = (compressedSize>>16) & 0xFF;
			out[16] = (compressedSize>>24) & 0xFF;
			out[17] = chunkData.length() & 0xFF;
			out[18] = (chunkData.length()>>8) & 0xFF;
			out[19] = (chunkData.length()>>16) & 0xFF;
			out[20] = (chunkData.length()>>24) & 0xFF;
			fwrite(out, 1, 9, file);

			//wait! write the palette first!
			if(mode == MODE_8BIT_PALETTE || mode == MODE_8BIT_PALETTE_ALPHA)
			{
				char* pout = new char[13 + (palette.size()*3)];
				char* pdeflatedData = new char[1 + (palette.size()*3)];
				unsigned int pcompressedSize = 0;
				pout[12] = static_cast<unsigned char> (palette.size());
				for(unsigned int i = 0;i < palette.size();i++)
				{
					pout[(i*3)+13] = palette[i].r;
					pout[(i*3)+14] = palette[i].g;
					pout[(i*3)+15] = palette[i].b;
				}
				deflateChunk(pout+12, 1 + (palette.size()*3), pdeflatedData, pcompressedSize);
				//header
				pout[0] = 'P'; pout[1] = 'L'; pout[2] = 'T'; pout[3] = 'E';
				pout[4] = pcompressedSize & 0xFF;
				pout[5] = (pcompressedSize>>8) & 0xFF;
				pout[6] = (pcompressedSize>>16) & 0xFF;
				pout[7] = (pcompressedSize>>24) & 0xFF;
				pout[8] = (1 + (palette.size()*3)) & 0xFF;
				pout[9] = ((1 + (palette.size()*3))>>8) & 0xFF;
				pout[10] = ((1 + (palette.size()*3))>>16) & 0xFF;
				pout[11] = ((1 + (palette.size()*3))>>24) & 0xFF;
				if(pcompressedSize >= 1 + (palette.size()*3))
				{
					pout[4] = pout[8];
					pout[5] = pout[9];
					pout[6] = pout[10];
					pout[7] = pout[11];
				}
				fwrite(pout, 1, 12, file);
				if(pcompressedSize >= 1 + (palette.size()*3))
					fwrite(pout+12, 1, 1 + (palette.size()*3), file);
				else
					fwrite(pdeflatedData, 1, 1 + (palette.size()*3), file);
				delete[] pout;
				delete[] pdeflatedData;
			}

			fwrite(out+9, 1, 12, file);
			//fwrite(chunkData.c_str(), 1, chunkData.length(), file);
			fwrite(deflatedData, 1, compressedSize, file);
			delete[] deflatedData;
			delete[] out;
		}

		unsigned short			width;
		unsigned short			height;
		unsigned char			mode;
		unsigned int*			columnMap;
		std::vector<Column *>	columns;
		std::vector<Pixel>		palette;
};

/// Standard rectangular image.
class BitmapImage
{
	public:
		BitmapImage(unsigned int width, unsigned int height, Pixel* data) : width(width), height(height), bitmap(data)
		{
		}
		~BitmapImage()
		{
			if(bitmap != NULL)
				delete[] bitmap;
		}

		/**
		 * Returns 8, 16, or 24.  Use with RequiresAlpha to determine 16, 24, and 32.
		 * @see RequiresAlpha
		 * @see IsGrayScale
		 */
		unsigned char GetMinimumColorDepth()
		{
			std::vector<Pixel> colors;
			for(unsigned int i = 0;i < width*height;i++)
			{
				bool found = false;
				for(unsigned int j = 0;j < colors.size();j++)
				{
					if(bitmap[i] == colors[j])
					{
						found = true;
						break;
					}
				}
				if(!found)
					colors.push_back(Pixel(bitmap[i]));
				if(colors.size() > 256)
					break;
			}
			if(colors.size() <= 256)
				return 8;
			for(unsigned int i = 0;i < colors.size();i++)
			{
				bool r5 = false;
				bool g5 = false; //green can be a 5 or 6 bit color
				bool g6 = false;
				bool b5 = false;
				for(unsigned int j = 0;j < 32;j++)
				{
					r5 = r5 || colors[i].r == ColorTable5Bit[j] ? true : false;
					g5 = g5 || colors[i].r == ColorTable5Bit[j] ? true : false;
					g6 = g6 || colors[i].r == ColorTable6Bit[(2*j)] || colors[i].r == ColorTable6Bit[(2*j)+1] ? true : false;
					b5 = b5 || colors[i].r == ColorTable5Bit[j] ? true : false;
				}
				if(!r5 || (!g5 && !g6) || !b5)
					return 24;
			}
			return 16;
		}
		bool IsGrayScale()
		{
			for(unsigned int i = 0;i < width*height;i++)
			{
				if(bitmap[i].r != bitmap[i].g || bitmap[i].g != bitmap[i].b)
					return false;
			}
			return true;
		}
		/// Polygonal Images always allow transparent and opaque, this returns if translucency is needed.
		bool RequiresAlpha()
		{
			for(unsigned int i = 0;i < width*height;i++)
			{
				if(bitmap[i].a != 0x00 && bitmap[i].a != 0xFF)
					return true;
			}
			return false;
		}

		unsigned int	width;
		unsigned int	height;
		Pixel*			bitmap;
};

void Error(const char* error)
{
	printf(error);
	exit(1);
}

bool noreduction = false;
int main(int argc, char* args[])
{
	if(argc < 2)
		Error("Usage: png2pgf <input> <output> [--noreduction]\n");

	for(int i = 0;i < argc;i++)
	{
		if(strcasecmp(args[i], "--noreduction") == 0)
		{
			noreduction = true;
		}
	}

	// Open the file for reading
	FILE *bmpFile = fopen(args[1], "rb");
	if(!bmpFile)
		Error("Could not open input.\n");
	printf("Input file: %s\n", args[1]);

	char header[14];
	fread(header, 1, 14, bmpFile);
	if(header[0] != 0x42 || header[1] != 0x4D)
		Error("Not a valid bitmap.\n");
	unsigned int size = READINT32(&header[2]);
	unsigned int offset = READINT32(&header[10]);

	char info[40];
	fread(info, 1, 40, bmpFile);
	unsigned int width = READINT32(&info[4]);
	unsigned int height = READINT32(&info[8]);
	printf("Size: %ux%u\n", width, height);
	unsigned short depth = READINT16(&info[14]);
	if(depth != 32 && depth != 24 && depth != 16 && depth != 8)
		Error("Must be a 32-bit, 24-bit, 16-bit, or 8-bit BMP.\n");
	printf("Depth: %d-bit\n", depth);
	if(READINT32(&info[16]) != 0)
		Error("This program doesn't support compression.\n");
	unsigned int rawSize = READINT32(&info[20]);
	unsigned int palSize = READINT32(&info[32]);
	palSize = palSize == 0 ? static_cast<unsigned int> (pow(2, depth)) : palSize;

	BitmapImage *image = new BitmapImage(width, height, new Pixel[width*height]);
	Pixel* palette = new Pixel[256];
	char palData[256*3];
	if(depth <= 8)
	{
		fread(palData, 1, palSize*3, bmpFile);
		for(unsigned int i = 0;i < palSize;i++)
		{
			palette[i] = Pixel(0, palData[(i*3)], palData[(i*3)+1], palData[(i*3)+2]);
		}
	}
	fseek(bmpFile, offset, SEEK_SET);
	for(int y = 0;y < height;y++)
	{
		for(int x = 0;x < width;x++)
		{
			if(depth == 32)
			{
				char tmp[4];
				fread(tmp, 1, 4, bmpFile);
				image->bitmap[(y*width)+x].r = tmp[0];
				image->bitmap[(y*width)+x].g = tmp[1];
				image->bitmap[(y*width)+x].b = tmp[2];
				image->bitmap[(y*width)+x].a = tmp[3];
			}
			else if(depth == 24)
			{
				char tmp[3];
				fread(tmp, 1, 3, bmpFile);
				image->bitmap[(y*width)+x].r = tmp[0];
				image->bitmap[(y*width)+x].g = tmp[1];
				image->bitmap[(y*width)+x].b = tmp[2];
				image->bitmap[(y*width)+x].a = 0xFF;
			}
			else if(depth == 16)
			{
				char tmp[2];
				fread(tmp, 1, 2, bmpFile);
				// Here's the deal, BMPs store 15-bit color XRRRRRGGGGGBBBBB
				// We store 16-bit RRRRRGGGGGGBBBBB, but in this code we store
				// 32-bit RGBA.  What we're going to do is convert the 15-bit 
				// to 32, then back down to 16 but in the form of 
				// RRRRRGGGGGXBBBBB
				image->bitmap[(y*width)+x].r = ColorTable5Bit[tmp[0] & 0x1F];
				image->bitmap[(y*width)+x].g = ColorTable5Bit[(((((unsigned int) static_cast<unsigned char> (tmp[0])) | ((unsigned int) static_cast<unsigned char> (tmp[1])<<8)) & 0x3E0) >> 5)];
				image->bitmap[(y*width)+x].b = ColorTable5Bit[(tmp[1] & 0x7C)>>2];
				image->bitmap[(y*width)+x].a = 0xFF;
			}
			else
			{
				char tmp[1];
				fread(tmp, 1, 1, bmpFile);
				image->bitmap[(y*width)+x] = Pixel(palette[tmp[1]]);
			}
		}
		int x = 0;
	}
	delete[] palette;
	fclose(bmpFile);

	printf("\nRequired color depth: Wait. . .");
	int requiredDepth = !noreduction ? image->GetMinimumColorDepth() : depth == 32 ? 24 : depth;
	bool grayScale = image->IsGrayScale();
	bool alphaChannel = !noreduction ? image->RequiresAlpha() : depth == 32;
	printf("\b\b\b\b\b\b\b\b\b%d        \n", requiredDepth);
	if(grayScale)
		printf("Gray Scale: yes\n");
	else
		printf("Gray Scale: no\n");
	if(alphaChannel)
		printf("Alpha Channel: yes\n\n");
	else
		printf("Alpha Channel: no\n\n");

	PolygonalImage img(width, height, CalculateMode(requiredDepth, grayScale, alphaChannel));
	for(int i = 0;i < width;i++)
	{
		Pixel* column = new Pixel[height];
		for(int j = 0;j < height;j++)
		{
			column[height - j - 1] = image->bitmap[(j*width)+i];
		}
		Column *clmn = Column::CreateColumn(column, height, CalculateMode(requiredDepth, grayScale, alphaChannel));
		img.columns.push_back(clmn);
	}
	FILE *pgfFile = fopen(args[2], "wb");
	img.Write(pgfFile);
	fclose(pgfFile);
	return 0;
}
